from Agent import *
from Tools import *
import parameters as pmt

class ModelSwarm(object):
	'''Questa classe definisce le funzionalità del modello.'''

	# inizializza l'ambiente
	def __init__(self, nFirms, nHouseholds, nCycles):
		self.nFirms = nFirms
		self.nHouseholds = nHouseholds
		self.firmsDict = {}
		self.householdsDict = {}
		self.PA = {}     
		self.idFirm = 2
		self.nu = 0

		self.marketPrice = 0
		self.nTotWorkers = 0
		self.nUnemployed = 0
		self.nNewFirms = 0
		self.totProfit = 0
		self.demand = []
		self.totalSavings = 0
		self.publicDebt = 0
		self.GDP = 0
		self.percDeficit = 0

		self.nCycles = nCycles
		self.t = 0
		self.conclude = False

		# quantità da osservare
		self.observer = {'time':[], 'nFirms':[], 'nWorkers':[], 'nNotWorkers':[], 'marketPrice':[], 'newFirms':[], 'totalProfit':[], 'percDeficit':[]}

	# descrive i parametri iniziali del modello
	def describeModel(self):
		print("Simulation starting at t=%d with:" % (self.t))
		print(self.nFirms, "firms")
		print(self.nHouseholds, "households")
		print("Unemployment rate ≃", pmt.unemploymentRate)
		print()

	# inizializza gli agenti
	def buildObjects(self):
		self.nTotWorkers = round(self.nHouseholds * (1 - pmt.unemploymentRate))
		# crea PubAdm
		workersPubAdm = round(self.nTotWorkers * pmt.publicEmploymentRate)
		self.PA = PubAdm(workersPubAdm)
		hList = []
		for i in range(workersPubAdm):
			hList.append(Household(is_employed=True))
		self.householdsDict[1] = hList  # l'azienda 1 equivale all'agente PA
		print("PubAdm created with", askAgent(self.PA, PubAdm.getNumWorkers), "workers")        
		# crea aziende
		self.nu = round((self.nTotWorkers/self.nFirms) * (1 - pmt.publicEmploymentRate)) # numero medio di lavoratori per azienda 
		self.nTotWorkers = 0 # lo riconto in un altro modo che dovrà restituire stesso risultato
		for i in range(self.nFirms):
			aFirm = Firm(self.nu + random.randint(-10, 10)) # genero azienda
			self.firmsDict[self.idFirm] = aFirm # aggiungo azienda al dizionario delle aziende, con rispettivo identificatore
			print("Firm num.", self.idFirm, "created with", askAgent(aFirm, Firm.getNumWorkers), "workers")
			self.nTotWorkers += askAgent(aFirm, Firm.getNumWorkers)
			# assume dipendenti
			hList = []
			for j in range(askAgent(aFirm, Firm.getNumWorkers)):
				hList.append(Household(is_employed=True))
			self.householdsDict[self.idFirm] = hList # associa ai lavoratori impiegati l'identificatore dell'azienda
			self.idFirm += 1 # cambia identificatore
		self.nTotWorkers += workersPubAdm
		# crea disoccupati
		self.nUnemployed = self.nHouseholds - self.nTotWorkers       
		hList = []
		for i in range(self.nUnemployed):
			hList.append(Household(is_employed=False))
		self.householdsDict[0] = hList # l'azienda 0 equivale alla "disoccupazione"

	# definisce le azioni

	# update observer
	def updateObserver(self):
		self.observer['time'].append(self.t)
		self.observer['nFirms'].append(self.nFirms)
		self.observer['nWorkers'].append(self.nTotWorkers)
		self.observer['nNotWorkers'].append(self.nUnemployed)
		self.observer['marketPrice'].append(self.marketPrice)
		self.observer['newFirms'].append(self.nNewFirms)
		self.observer['totalProfit'].append(self.totProfit)
		self.observer['percDeficit'].append(self.percDeficit)

	# calcola il prezzo dato dall'intersezione di domanda e offerta
	def setMarketPrice(self, demand, supply):
		self.marketPrice = float(demand) / supply

	# controlla lo stato di salute delle aziende e decide quante falliscono e quante nuove ne nascono
	def checkFirmsHealth(self):
		idxToDel = []
		newFirms = []
		for key, aFirm in self.firmsDict.items():
			if askAgent(aFirm, Firm.checkHealth) == -1:
				idxToDel.append(key)
			if askAgent(aFirm, Firm.checkHealth) == 1:
				newFirms.append(key)
		return (idxToDel, newFirms)

	# rimuove aziende fallite e mette in disoccupazione i lavoratori
	def removeFirms(self, idxToDel):
		self.firmsDict = {key:self.firmsDict[key] for key in list(self.firmsDict.keys()) if key not in idxToDel} # rimuovo aziende
		self.nFirms = len(list(self.firmsDict.keys())) # aggiorno numero aziende
		print()
		#print(len(idxToDel), " firms removed")

		#print("unemployed workers:", self.nUnemployed)
		for key in idxToDel:
			for worker in self.householdsDict[key]: # aggiungo lavoratori disoccupati all'azienda 0
				askAgent(worker, Household.setEmploymentStatus, False)
				askAgent(worker, Household.setInvestmentStatus, False)
				self.householdsDict[0].append(worker)
				self.nUnemployed += 1
				self.nTotWorkers -= 1
			del[self.householdsDict[key]] # elimino azienda key dalla lista delle households

	# genera una nuova azienda
	def addFirms(self, newFirms):
		self.nNewFirms = 0
		for key in newFirms:
			# la nuova azienda ha lo stesso numero di lavoratori di quella vecchia
			#nNewWorkers = askAgent(self.firmsDict[key], Firm.getNumWorkers)
			nNewWorkers = self.nu + random.randint(-10, 10)

			if nNewWorkers <= self.nUnemployed: # controllo che ci sia abbastanza forza lavoro
				newFirm = Firm(nNewWorkers) # genero azienda
				self.nNewFirms += 1
				self.idFirm += 1 # associo un identificativo alla nuova azienda
				self.firmsDict[self.idFirm] = newFirm
				self.nFirms += 1

				# modifico percentuale occupazione
				self.nTotWorkers += nNewWorkers # aggiorno contatore occupati
				self.householdsDict[self.idFirm] = self.householdsDict[0][:nNewWorkers] # aggiungo lavoratori alla nuova azienda
				for newWorker in self.householdsDict[self.idFirm]:
					askAgent(newWorker, Household.setEmploymentStatus, True)
					self.nUnemployed -= 1
				self.householdsDict[0] = self.householdsDict[0][nNewWorkers:] # tolgo gli stessi lavoratori da azienda 0

				# l'azienda che investe deve ripagare il debito
				askAgent(self.firmsDict[key], Firm.setInvestmentStatus, nNewWorkers)

	# assume/licenzia lavoratori
	def planProductionAndHireFire(self):
		for key, aFirm in self.firmsDict.items():
			extraWorkforce = aFirm.planProduction(previousDemand=self.demand[self.t - 1], actualDemand=self.demand[self.t])

			# assunzioni
			if extraWorkforce > 0:
				for i in range(extraWorkforce):
					if self.nUnemployed > 0:
						newWorker = self.householdsDict[0][0]
						askAgent(newWorker, Household.setEmploymentStatus, True)
						self.householdsDict[key].append(newWorker)
						del(self.householdsDict[0][0])
						askAgent(aFirm, Firm.updateNumWorkers, 1)
						self.nTotWorkers += 1
						self.nUnemployed -= 1

			# licenziamenti
			if extraWorkforce < 0:
				for i in range(extraWorkforce):
					newUnemployed = self.householdsDict[key][0]
					askAgent(newUnemployed, Household.setEmploymentStatus, False)
					askAgent(newUnemployed, Household.setInvestmentStatus, False)
					self.householdsDict[0].append(newUnemployed)
					del(self.householdsDict[key][0])
					askAgent(aFirm, Firm.updateNumWorkers, -1)
					self.nTotWorkers -= 1
					self.nUnemployed += 1

	# calcolo GDP                    
	def setGDP(self, demand, PublicExpenditure):
		self.GDP = (demand + PublicExpenditure)/(1 - pmt.tradeBalance)
       
    
	# simulazione
	def run(self):

		while not self.conclude:
			print()
			print("Step t =", self.t)
			print("Workers unemployed:", self.nUnemployed)
			print("Workers employed:", self.nTotWorkers, ";", self.nTotWorkers - askAgent(self.PA, PubAdm.getNumWorkers), "for private and ", askAgent(self.PA, PubAdm.getNumWorkers), "for public")
			print("Firms:", self.nFirms)

			# calcolo offerta
			supply = askEachAgentIn(list(self.firmsDict.values()), Firm.calcProduction)

			# calcolo domanda
			demandHouseholds = askEachAgentIn(list(self.householdsDict.values()), Household.calcConsumption)
			demandFirms = askEachAgentIn(list(self.firmsDict.values()), Firm.calcConsumption)
			demandPubAdm = askAgent(self.PA, PubAdm.calcConsumption)
			self.demand.append(demandHouseholds + demandFirms + demandPubAdm)
            
			# calcolo prezzo di mercato
			self.setMarketPrice(demand=self.demand[self.t], supply=supply)

			# setta prezzo aziende
			askEachAgentIn(list(self.firmsDict.values()), Firm.setFirmPrice, self.marketPrice)

			# calcola profitto aziende (e aggiorna stato salute di conseguenza)
			self.totProfit = askEachAgentIn(list(self.firmsDict.values()), Firm.calcProfit)

			# controlla stato salute
			idxToDel, newFirms = self.checkFirmsHealth()
			#print("Firms in bankrupt:", len(idxToDel))
			#print("New firms to be created:", len(newFirms))

			# rimuove aziende fallite
			self.removeFirms(idxToDel)

			# genera nuove aziende
			self.addFirms(newFirms)

			# pianifica produzione al tempo successivo, per t > 0
			if self.t > 0:
				# le aziende assumono/licenziano a seconda della produzione pianificata
				self.planProductionAndHireFire()
                
			# calcolo spesa pubblica, risparmi, GDP e deficit.
			PublicExpenditure = askAgent(self.PA, PubAdm.calcPublicExpenditure)
			self.setGDP(demand = demandHouseholds + demandFirms, PublicExpenditure = PublicExpenditure)
            
			self.totalSavings = askEachAgentIn(list(self.householdsDict.values()), Household.calcSavings)
			deficAndMaas = askAgent(self.PA, PubAdm.calcDeficit, self.totalSavings, self.GDP)           
			deficit = deficAndMaas[0]
			self.publicDebt += deficit
			self.percDeficit = deficAndMaas[1]

			print("Total supply:", supply)
			print("Total demand:", self.demand[self.t])
			print("Market price:", self.marketPrice)
			print("PublicExpenditure:", PublicExpenditure)
			print("GDP:", self.GDP)
			print("Total savings:", self.totalSavings)
			print("Actual deficit:", deficit)
			print("Public debt:", self.publicDebt)
			print("Percentage Maastricht:", self.percDeficit, "%") 
            
			# update observer
			self.updateObserver()            

			# check if all firms went bankrupt!
			if self.nFirms == 0:
				print("All the firms went bankrupt! :(")
				return self.observer

			# update time
			self.t += 1
			if self.t == self.nCycles:
				self.conclude = True

		return self.observer